"""Provides a client for the Halo Infinite API.

Endpoints are documented at:
https://settings.svc.halowaypoint.com/settings/hipc/e2a0a7c6-6efe-42af-9283-c2ab73250c48

Additionally inspected the network traffic while navigating Halo Waypoint.
"""

from typing import Iterable, Literal
from uuid import UUID

from aiohttp import ClientResponse, ClientSession
from aiolimiter import AsyncLimiter

from .xuid import wrap_xuid

GAMECMS_HACS_HOST = "https://gamecms-hacs.svc.halowaypoint.com"
SKILL_HOST = "https://skill.svc.halowaypoint.com:443"
STATS_HOST = "https://halostats.svc.halowaypoint.com:443"
UGC_DISCOVERY_HOST = "https://discovery-infiniteugc.svc.halowaypoint.com:443"


def _create_limiter(rate_per_second: int) -> AsyncLimiter:
    """Return an AsyncLimiter with the given rate per second."""
    # Setting the max rate to 1 disallows bursts.
    return AsyncLimiter(1, 1 / rate_per_second)


class HaloInfiniteClient:
    """A client for the Halo Infinite API."""

    def __init__(
        self,
        session: ClientSession,
        spartan_token: str,
        clearance_token: str,
        requests_per_second: int | None = 5,
    ) -> None:
        """Initialize a client for the Halo Infinite API.

        Raw `aiohttp` `ClientResponses` are returned from each method. The
        caller is responsible for handling the response via custom parsing or by
        using one of the provided parsers from the `spnkr.parsers` module.

        Args:
            session: The aiohttp session to use.
            spartan_token: The spartan token used to authenticate with the API.
            clearance_token: The clearance token used to authenticate with the API.
            requests_per_second: The rate limit to use. Defaults to 5 requests per second.
                Set to None to disable rate limiting.
        """
        self._session = session
        headers = {
            "Accept": "application/json",
            "x-343-authorization-spartan": spartan_token,
            "343-clearance": clearance_token,
        }
        self._session.headers.update(headers)

        self._rate_limiter = None
        if requests_per_second is not None:
            self._rate_limiter = _create_limiter(requests_per_second)

    async def _get(self, url: str, **kwargs) -> ClientResponse:
        """Make a GET request to the given URL."""
        if self._rate_limiter is None:
            return await self._session.get(url, **kwargs)
        async with self._rate_limiter:
            return await self._session.get(url, **kwargs)

    async def get_medal_metadata(self) -> ClientResponse:
        """Get details for all medals obtainable in the game.

        Parsers:
            - [MedalMetadata][spnkr.parsers.pydantic.gamecms_hacs.MedalMetadata]
            - [parse_medal_metadata][spnkr.parsers.records.gamecms_hacs.parse_medal_metadata]

        Returns:
            The medal metadata.
        """
        url = f"{GAMECMS_HACS_HOST}/hi/Waypoint/file/medals/metadata.json"
        return await self._get(url)

    async def get_match_skill(
        self, match_id: str | UUID, xuids: Iterable[str | int]
    ) -> ClientResponse:
        """Get player CSR and team MMR values for a given match and player list.

        Args:
            match_id: Halo Infinite match ID.
            xuids: The Xbox Live IDs of the match's players. Only
                players in this list will have their skill data returned.

        Parsers:
            - [MatchSkill][spnkr.parsers.pydantic.skill.MatchSkill]
            - [parse_match_skill][spnkr.parsers.records.skill.parse_match_skill]

        Returns:
            The skill data for the match.
        """
        url = f"{SKILL_HOST}/hi/matches/{match_id}/skill"
        params = {"players": [wrap_xuid(x) for x in xuids]}
        return await self._get(url, params=params)

    async def get_playlist_csr(
        self, playlist_id: str | UUID, xuids: Iterable[str | int]
    ) -> ClientResponse:
        """Get player CSR values for a given playlist and player list.

        Args:
            playlist_id: Halo Infinite playlist asset ID.
            xuids: The Xbox Live IDs of the players.

        Parsers:
            - [PlaylistCsr][spnkr.parsers.pydantic.skill.PlaylistCsr]
            - [parse_playlist_csr][spnkr.parsers.records.skill.parse_playlist_csr]

        Returns:
            The summary CSR data for the players in the given playlist.
        """
        url = f"{SKILL_HOST}/hi/playlist/{playlist_id}/csrs"
        params = {"players": [wrap_xuid(x) for x in xuids]}
        return await self._get(url, params=params)

    async def get_match_count(self, xuid: str | int) -> ClientResponse:
        """Get match counts across different game experiences for a player.

        The counts returned are for custom matches, matchmade matches, local
        matches, and total matches.

        Args:
            xuid: Xbox Live ID of the player to get counts for.

        Parsers:
            - [MatchCount][spnkr.parsers.pydantic.stats.MatchCount]
            - [parse_match_count][spnkr.parsers.records.stats.parse_match_count]

        Returns:
            The match counts.
        """
        url = f"{STATS_HOST}/hi/players/{wrap_xuid(xuid)}/matches/count"
        return await self._get(url)

    async def get_match_history(
        self,
        xuid: str | int,
        start: int = 0,
        count: int = 25,
        match_type: Literal["all", "matchmaking", "custom", "local"] = "all",
    ) -> ClientResponse:
        """Request a batch of matches from a player's match history.

        Args:
            xuid: Xbox Live ID of the player to request matches for.
            start: Index of the first match to request, starting at 0.
            count: The number of matches to request. Maximum number of results returned is 25.
            match_type: The type of matches to return. One of "all", "matchmaking", "custom", or "local".

        Parsers:
            - [MatchHistory][spnkr.parsers.pydantic.stats.MatchHistory]
            - [parse_match_history][spnkr.parsers.records.stats.parse_match_history]

        Returns:
            The requested match history "page" of results.
        """
        url = f"{STATS_HOST}/hi/players/{wrap_xuid(xuid)}/matches"
        params = {"start": start, "count": count, "type": match_type}
        return await self._get(url, params=params)

    async def get_match_stats(self, match_id: str | UUID) -> ClientResponse:
        """Request match details using the Halo Infinite match GUID.

        Args:
            match_id: Halo Infinite GUID identifying the match.

        Parsers:
            - [MatchStats][spnkr.parsers.pydantic.stats.MatchStats]
            - [parse_match_info][spnkr.parsers.records.stats.parse_match_info]
            - [parse_player_core_stats][spnkr.parsers.records.stats.parse_player_core_stats]
            - [parse_player_medals][spnkr.parsers.records.stats.parse_player_medals]
            - [parse_team_core_stats][spnkr.parsers.records.stats.parse_team_core_stats]

        Returns:
            The match details.
        """
        url = f"{STATS_HOST}/hi/matches/{match_id}/stats"
        return await self._get(url)

    async def get_ugc_game_variant(
        self, asset_id: str | UUID, version_id: str | UUID
    ) -> ClientResponse:
        """Get details about a game mode.

        Args:
            asset_id: The asset ID of the game variant.
            version_id: The version ID of the game variant.

        Parsers:
            - [UgcGameVariant][spnkr.parsers.pydantic.ugc_discovery.UgcGameVariant]
            - [parse_asset][spnkr.parsers.records.ugc_discovery.parse_asset]

        Returns:
            The game variant details.
        """
        endpoint = f"/hi/ugcGameVariants/{asset_id}/versions/{version_id}"
        url = f"{UGC_DISCOVERY_HOST}{endpoint}"
        return await self._get(url)

    async def get_map_mode_pair(
        self,
        asset_id: str | UUID,
        version_id: str | UUID,
    ) -> ClientResponse:
        """Get details about a map mode pair.

        Args:
            asset_id: The asset ID of the map mode pair.
            version_id: The version ID of the map mode pair.

        Parsers:
            - [MapModePair][spnkr.parsers.pydantic.ugc_discovery.MapModePair]
            - [parse_asset][spnkr.parsers.records.ugc_discovery.parse_asset]

        Returns:
            The map mode pair details.
        """
        endpoint = f"/hi/mapModePairs/{asset_id}/versions/{version_id}"
        url = f"{UGC_DISCOVERY_HOST}{endpoint}"
        return await self._get(url)

    async def get_map(
        self, asset_id: str | UUID, version_id: str | UUID
    ) -> ClientResponse:
        """Get details about a map.

        Args:
            asset_id: The asset ID of the map.
            version_id: The version ID of the map.

        Parsers:
            - [Map][spnkr.parsers.pydantic.ugc_discovery.Map]
            - [parse_asset][spnkr.parsers.records.ugc_discovery.parse_asset]

        Returns:
            The map details.
        """
        endpoint = f"/hi/maps/{asset_id}/versions/{version_id}"
        url = f"{UGC_DISCOVERY_HOST}{endpoint}"
        return await self._get(url)

    async def get_playlist(
        self,
        asset_id: str | UUID,
        version_id: str | UUID,
    ) -> ClientResponse:
        """Get details about a playlist.

        Args:
            asset_id: The asset ID of the playlist.
            version_id: The version ID of the playlist.

        Parsers:
            - [Playlist][spnkr.parsers.pydantic.ugc_discovery.Playlist]
            - [parse_asset][spnkr.parsers.records.ugc_discovery.parse_asset]

        Returns:
            The playlist details.
        """
        endpoint = f"/hi/playlists/{asset_id}/versions/{version_id}"
        url = f"{UGC_DISCOVERY_HOST}{endpoint}"
        return await self._get(url)



"""Enumerated data types used by the Halo Infinite API."""

from enum import IntEnum, StrEnum


class AssetKind(IntEnum):
    """Types of assets used by Halo Infinite."""

    FILM = 1
    """A film asset."""
    MAP = 2
    """A map asset."""
    PLAYLIST = 3
    """A matchmaking playlist asset."""
    PREFAB = 4
    """A prefabricated object asset."""
    TEST_ASSET = 5
    """A test asset."""
    UGC_GAME_VARIANT = 6
    """A user-generated content game mode asset."""
    MAP_MODE_PAIR = 7
    """A map and game mode pair asset."""
    PROJECT = 8
    """A project asset."""
    MANIFEST = 9
    """A manifest asset."""
    ENGINE_GAME_VARIANT = 10
    """An engine game mode asset."""


class BotDifficulty(IntEnum):
    """Bot difficulty levels."""

    RECRUIT = 4
    """Recruit bots. Lowest difficulty."""
    MARINE = 1
    """Marine bots. 2nd lowest difficulty."""
    ODST = 2
    """ODST bots. 2nd highest difficulty."""
    SPARTAN = 3
    """Spartan bots. Highest difficulty."""


class GameVariantCategory(IntEnum):
    """Categories of multiplayer game modes."""

    SLAYER = 6
    """Slayer game modes."""
    ATTRITION = 7
    """Attrition game modes."""
    ELIMINATION = 8
    """Elimination game modes."""
    FIESTA = 9
    """Fiesta game modes."""
    STRONGHOLDS = 11
    """Strongholds game modes."""
    BASTION = 12
    """Bastion game modes."""
    TOTAL_CONTROL = 14
    """Total Control game modes."""
    CTF = 15
    """Capture the Flag game modes."""
    ASSAULT = 16
    """Assault game modes."""
    EXTRACTION = 17
    """Extraction game modes."""
    ODDBALL = 18
    """Oddball game modes."""
    STOCKPILE = 19
    """Stockpile game modes."""
    JUGGERNAUT = 20
    """Juggernaut game modes."""
    INFECTION = 22
    """Infection game modes."""
    ESCORT = 23
    """Escort game modes."""
    GUN_GAME = 24
    """Gun Game game modes."""
    GRIFBALL = 25
    """Grifball game modes."""
    TEST_ENGINE = 32
    """Test Engine game modes."""
    LAND_GRAB = 39
    """Land Grab game modes."""


class LifecycleMode(IntEnum):
    """General categories of game modes."""

    CUSTOM = 1
    """Custom games."""
    MATCHMADE = 3
    """Matchmade games."""


class MedalDifficulty(IntEnum):
    """Difficulty of medals obtainable in matchmaking."""

    # Values line up to indices in the medal metadata response content.
    NORMAL = 0
    """Normal difficulty medals. Easiest to obtain."""
    HEROIC = 1
    """Heroic difficulty medals. 2nd easiest to obtain."""
    LEGENDARY = 2
    """Legendary difficulty medals. 2nd hardest to obtain."""
    MYTHIC = 3
    """Mythic difficulty medals. Hardest to obtain."""


class MedalType(IntEnum):
    """Types of medals obtainable in matchmaking."""

    # Values line up to indices in the medal metadata response content.
    SPREE = 0
    """General or weapon-specific killing sprees."""
    MODE = 1
    """Game mode-specific medals."""
    MULTIKILL = 2
    """Kill multiple enemies in quick succession."""
    PROFICIENCY = 3
    """Medals related to player proficiency."""
    SKILL = 4
    """Medals awarded for skillful play."""
    STYLE = 5
    """Medals awarded for stylish play."""


class Outcome(IntEnum):
    """Match outcome options."""

    TIE = 1
    """Match ended in a tie."""
    WIN = 2
    """Match was won."""
    LOSS = 3
    """Match was lost."""
    DID_NOT_FINISH = 4
    """Match was not finished."""


class PlayerType(IntEnum):
    """Types of players."""

    HUMAN = 1
    """Human players."""
    BOT = 2
    """AI players."""


class PlaylistExperience(IntEnum):
    """General categories of playlists."""

    ARENA = 2
    """Arena playlists. Typically 4v4."""
    BIG_TEAM_BATTLE = 3
    """Big team battle playlists. Typically 12v12."""
    PVE = 4
    """Player vs. environment playlists."""
    FEATURED = 5
    """Featured playlists. Rotates frequently."""


class SkillResultCode(IntEnum):
    """Result codes for skill requests."""

    SUCCESS = 0
    """Skill request was successful."""


class SubTier(IntEnum):
    """Sub-tiers of skill rankings."""

    NOT_APPLICABLE = 0
    """Not applicable."""
    ONE = 1
    """1st sub-tier."""
    TWO = 2
    """2nd sub-tier."""
    THREE = 3
    """3rd sub-tier."""
    FOUR = 4
    """4th sub-tier."""
    FIVE = 5
    """5th sub-tier."""
    SIX = 6
    """6th sub-tier. Highest sub-tier before advancing to the next tier."""


class Tier(StrEnum):
    """Tiers of skill rankings."""

    NOT_APPLICABLE = ""
    """Not applicable."""
    BRONZE = "Bronze"
    """Bronze tier. Lowest tier."""
    SILVER = "Silver"
    """Silver tier."""
    GOLD = "Gold"
    """Gold tier."""
    PLATINUM = "Platinum"
    """Platinum tier."""
    DIAMOND = "Diamond"
    """Diamond tier."""
    ONYX = "Onyx"
    """Onyx tier. Highest tier."""



"""Test the SPNKR API client."""

import time
from unittest.mock import AsyncMock

import pytest

from spnkr.client import HaloInfiniteClient


class MockSession:
    def __init__(self) -> None:
        self.headers = {}
        self.get = AsyncMock()


SESSION = MockSession()


@pytest.fixture
def client():
    return HaloInfiniteClient(SESSION, "spartan", "clearance")  # type: ignore


def test_header_update(client: HaloInfiniteClient):
    """Test that the client headers are updated as expected."""
    assert client._session.headers == {
        "Accept": "application/json",
        "x-343-authorization-spartan": "spartan",
        "343-clearance": "clearance",
    }


def test_async_limiter_set(client: HaloInfiniteClient):
    """Test that the async limiter is set as expected."""
    assert client._rate_limiter is not None
    assert client._rate_limiter.max_rate / client._rate_limiter.time_period == 5


@pytest.mark.asyncio
async def test_rate_limiter(client: HaloInfiniteClient):
    """Test that the rate limiter limits requests as expected."""
    t0 = time.time()
    # Use 6 requests due to the default rate of 5 requests per second.
    for _ in range(6):
        await client._get("url")
    t1 = time.time()
    assert t1 - t0 >= 1


@pytest.mark.asyncio
async def test_rate_limiter_none(client: HaloInfiniteClient):
    """Test that the rate limiter does not limit requests if None."""
    client._rate_limiter = None
    t0 = time.time()
    for _ in range(6):
        await client._get("url")
    t1 = time.time()
    assert t1 - t0 < 1


@pytest.mark.asyncio
async def test_get_medal_metadata(client: HaloInfiniteClient):
    await client.get_medal_metadata()
    SESSION.get.assert_called_with(
        "https://gamecms-hacs.svc.halowaypoint.com/hi/Waypoint/file/medals/metadata.json"
    )


@pytest.mark.asyncio
async def test_get_map_mode_pair(client: HaloInfiniteClient):
    await client.get_map_mode_pair("asset_id", "version_id")
    SESSION.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/mapModePairs/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_playlist(client: HaloInfiniteClient):
    await client.get_playlist("asset_id", "version_id")
    SESSION.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/playlists/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_map(client: HaloInfiniteClient):
    await client.get_map("asset_id", "version_id")
    SESSION.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/maps/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_ugc_game_variant(client: HaloInfiniteClient):
    await client.get_ugc_game_variant("asset_id", "version_id")
    SESSION.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/ugcGameVariants/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_match_skill(client: HaloInfiniteClient):
    await client.get_match_skill("match_id", [123])
    SESSION.get.assert_called_with(
        "https://skill.svc.halowaypoint.com:443/hi/matches/match_id/skill",
        params={"players": ["xuid(123)"]},
    )


@pytest.mark.asyncio
async def test_get_playlist_csr(client: HaloInfiniteClient):
    await client.get_playlist_csr("playlist_id", [123, 456])
    SESSION.get.assert_called_with(
        "https://skill.svc.halowaypoint.com:443/hi/playlist/playlist_id/csrs",
        params={"players": ["xuid(123)", "xuid(456)"]},
    )


@pytest.mark.asyncio
async def test_get_match_count(client: HaloInfiniteClient):
    await client.get_match_count(123)
    SESSION.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/players/xuid(123)/matches/count"
    )


@pytest.mark.asyncio
async def test_get_match_history(client: HaloInfiniteClient):
    await client.get_match_history(123, 0, 10, "all")
    SESSION.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/players/xuid(123)/matches",
        params={"start": 0, "count": 10, "type": "all"},
    )


@pytest.mark.asyncio
async def test_get_match_stats(client: HaloInfiniteClient):
    await client.get_match_stats("match_id")
    SESSION.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/matches/match_id/stats"
    )


if __name__ == "__main__":
    pytest.main()



"""Convert Xbox Live IDs (XUID) to and from strings and integers."""

import re

_PATTERN = re.compile(r"^xuid\((\d+)\)$")


def wrap_xuid(value: str | int) -> str:
    """Wrap a XUID value in the "xuid()" format.

    Args:
        value: The XUID value to wrap.

    Returns:
        The wrapped XUID value as a string.
    """
    value = str(value)
    if value[0] == "x":
        return value
    return f"xuid({value})"


def unwrap_xuid(value: str | int) -> int:
    """Unwrap a XUID value from the "xuid()" format.

    Args:
        value: The XUID value to unwrap.

    Raises:
        ValueError: If the XUID value is invalid.

    Returns:
        The unwrapped XUID value as an integer.
    """
    if isinstance(value, int) or value.isdigit():
        return int(value)

    match = _PATTERN.match(value)
    if not match:
        raise ValueError(f"Invalid XUID: {value!r}")
    return int(match.group(1))



"""Test the spnkr.xuid module."""

import pytest

from spnkr.xuid import unwrap_xuid, wrap_xuid

WRAPPABLE = [
    ("1234567890", "xuid(1234567890)"),
    ("xuid(1234567890)", "xuid(1234567890)"),
    (1234567890, "xuid(1234567890)"),
]
UNWRAPPABLE = [
    ("1234567890", 1234567890),
    ("xuid(1234567890)", 1234567890),
    (1234567890, 1234567890),
]
INVALID = ["1234y567890", "(123)", "xuid(1234567890x)", "bid(1234567890)"]


@pytest.mark.parametrize("value, expected", WRAPPABLE)
def test_wrap_xuid(value: str | int, expected: str):
    """Test that the XUID is wrapped."""
    assert wrap_xuid(value) == expected


@pytest.mark.parametrize("value, expected", UNWRAPPABLE)
def test_unwrap_xuid(value: str | int, expected: int):
    """Test that the XUID is unwrapped."""
    assert unwrap_xuid(value) == expected


@pytest.mark.parametrize("value", INVALID)
def test_unwrap_xuid_invalid(value: str | int):
    """Test that invalid XUIDs raise an error."""
    with pytest.raises(ValueError):
        unwrap_xuid(value)



"""Tools for Halo Infinite data analysis."""

import math
from enum import IntEnum
from typing import NamedTuple

from .parsers.refdata import (
    BotDifficulty,
    GameVariantCategory,
    LifecycleMode,
    MedalDifficulty,
    MedalType,
    Outcome,
    PlayerType,
    PlaylistExperience,
    SubTier,
    Tier,
)
from .xuid import unwrap_xuid, wrap_xuid

__all__ = [
    "wrap_xuid",
    "unwrap_xuid",
    "BOT_DIFFICULTY_MAP",
    "GAME_CATEGORY_MAP",
    "LIFECYCLE_MAP",
    "MEDAL_DIFFICULTY_MAP",
    "MEDAL_TYPE_MAP",
    "OUTCOME_MAP",
    "PLAYER_TYPE_MAP",
    "PLAYLIST_EXPERIENCE_MAP",
    "get_rank_from_csr",
    "Rank",
]


def _intenum_to_mapping(enum: type[IntEnum]) -> dict[int, str]:
    """Create a dict of value to name for an `IntEnum`."""
    return {e.value: e.name for e in enum}


BOT_DIFFICULTY_MAP: dict[int, str] = _intenum_to_mapping(BotDifficulty)
"""Mapping of bot difficulty values to names."""
GAME_CATEGORY_MAP: dict[int, str] = _intenum_to_mapping(GameVariantCategory)
"""Mapping of game variant category values to names."""
LIFECYCLE_MAP: dict[int, str] = _intenum_to_mapping(LifecycleMode)
"""Mapping of lifecycle mode values to names."""
MEDAL_DIFFICULTY_MAP: dict[int, str] = _intenum_to_mapping(MedalDifficulty)
"""Mapping of medal difficulty values to names."""
MEDAL_TYPE_MAP: dict[int, str] = _intenum_to_mapping(MedalType)
"""Mapping of medal type values to names."""
OUTCOME_MAP: dict[int, str] = _intenum_to_mapping(Outcome)
"""Mapping of match outcome values to names."""
PLAYER_TYPE_MAP: dict[int, str] = _intenum_to_mapping(PlayerType)
"""Mapping of player type values to names."""
PLAYLIST_EXPERIENCE_MAP: dict[int, str] = _intenum_to_mapping(
    PlaylistExperience
)
"""Mapping of playlist experience values to names."""


class Rank(NamedTuple):
    """A CSR tier and sub-tier.

    Attributes:
        tier: CSR tier.
        sub_tier: CSR sub-tier.
    """

    tier: Tier
    sub_tier: SubTier

    def __str__(self) -> str:
        if self.tier is Tier.ONYX:
            return self.tier.value
        return f"{self.tier.value} {self.sub_tier.value}"


def get_rank_from_csr(csr: int | float) -> Rank:
    """Get the tier and sub-tier from a CSR value.

    Args:
        csr: CSR value to convert.

    Returns:
        A ranking named tuple containing `tier` and `subtier`.
    """
    csr = math.floor(csr)
    quotient, remainder = divmod(csr, 300)
    if quotient == 0:
        tier = Tier.BRONZE
    elif quotient == 1:
        tier = Tier.SILVER
    elif quotient == 2:
        tier = Tier.GOLD
    elif quotient == 3:
        tier = Tier.PLATINUM
    elif quotient == 4:
        tier = Tier.DIAMOND
    else:
        # Onyx doesn't have sub-tiers
        return Rank(Tier.ONYX, SubTier.NOT_APPLICABLE)
    sub_tier = SubTier((remainder // 50) + 1)
    return Rank(tier, sub_tier)
